@@ -265,6 +265,7 @@ impl Example {
             },
             depth_stencil: None,
             multisample: wgpu::MultisampleState::default(),
+            multiview: None,
         });
 
         // Done
@@ -353,6 +354,7 @@ fn main() {
     let adapter = block_on(instance.request_adapter(&wgpu::RequestAdapterOptions {
         power_preference: wgpu::PowerPreference::HighPerformance,
         compatible_surface: Some(&surface),
+        force_fallback_adapter: false,
     }))
     .unwrap();
 
@@ -481,7 +483,7 @@ fn main() {
                 imgui.io_mut().update_delta_time(now - last_frame);
                 last_frame = now;
 
-                let frame = match surface.get_current_frame() {
+                let frame = match surface.get_current_texture() {
                     Ok(frame) => frame,
                     Err(e) => {
                         eprintln!("dropped frame: {:?}", e);
@@ -494,7 +496,6 @@ fn main() {
                 let ui = imgui.frame();
 
                 let view = frame
-                    .output
                     .texture
                     .create_view(&wgpu::TextureViewDescriptor::default());
 
@@ -537,7 +538,7 @@ fn main() {
                     // Only render example to example_texture if thw window is not collapsed
                     example.setup_camera(&queue, size);
                     example.render(
-                        &renderer.textures.get(example_texture_id).unwrap().view(),
+                        renderer.textures.get(example_texture_id).unwrap().view(),
                         &device,
                         &queue,
                     );
@@ -572,6 +573,7 @@ fn main() {
                 drop(rpass);
 
                 queue.submit(Some(encoder.finish()));
+                frame.present();
             }
             _ => (),
         }
